// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: GPL-3.0-or-later

use {
    libc::c_char,
    nix::errno::Errno,
    once_cell::sync::Lazy,
    std::ffi::CString,
    std::mem::{align_of, size_of},
    trust_dns_proto::rr::resource::Record,
};

static RUNTIME: Lazy<std::io::Result<tokio::runtime::Runtime>> = Lazy::new(|| {
    // The runtime should remain single-threaded, some
    // programs depend on it (e.g. programs calling unshare())
    let rt = tokio::runtime::Builder::new_current_thread()
        .enable_time()
        .enable_io()
        .build()?;
    Ok(rt)
});

pub fn write_record_name_to_buf<'a>(
    record: &Record,
    buf: &'a mut [u8],
) -> std::io::Result<(&'a mut [u8], *mut c_char)> {
    use std::io::{Error, ErrorKind};
    let name = CString::new(record.name().to_utf8())
        .map_err(|e| Error::new(ErrorKind::InvalidData, e.to_string()))?;

    let offset = buf.as_ptr().align_offset(align_of::<libc::c_char>());
    let name_src = name.as_bytes_with_nul();
    let l = name_src.len();
    if buf.len() < offset + l {
        return Err(Errno::ERANGE.into());
    }

    let name_dest = unsafe { buf.as_mut_ptr().add(offset) };
    unsafe {
        std::ptr::copy_nonoverlapping(name_src.as_ptr(), name_dest, l);
    }
    Ok((&mut buf[(offset + l)..], name_dest as *mut c_char))
}

pub fn write_vector_to_buf<'a, T>(
    buf: &'a mut [u8],
    v: Vec<T>,
) -> std::io::Result<(&'a mut [u8], *mut T)> {
    let offset = buf.as_ptr().align_offset(align_of::<*mut T>());
    let l = size_of::<*mut c_char>() * v.len();
    if buf.len() < offset + l {
        return Err(Errno::ERANGE.into());
    }

    let buf = &mut buf[offset..];
    let dest = buf.as_mut_ptr().cast::<T>();
    unsafe {
        std::ptr::copy_nonoverlapping(v.as_ptr() as *const T, dest, l);
    }
    Ok((&mut buf[l..], dest))
}

pub fn set_if_valid<T>(ptr: *mut T, val: T) {
    if !ptr.is_null() {
        unsafe { *ptr = val };
    }
}

pub fn block_on<F>(f: F) -> anyhow::Result<F::Output>
where
    F: std::future::Future,
{
    // TODO: use std::panic::catch_unwind() to be resilient
    // to OOM problems

    use std::ops::Deref;
    match RUNTIME.deref() {
        Ok(rt) => Ok(rt.block_on(async { f.await })),
        Err(e) => anyhow::bail!("Unable to start tokio runtime: {}", e),
    }
}
