// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: GPL-3.0-or-later

use {
    nix::unistd::Uid,
    serde::{Deserialize, Serialize},
    std::net::IpAddr,
    zbus::{dbus_proxy, Result},
};

#[dbus_proxy(
    default_service = "com.endlessm.ParentalControls",
    interface = "com.endlessm.ParentalControls.Dns",
    default_path = "/com/endlessm/ParentalControls/Dns",
    gen_blocking = false
)]
trait MalcontentDns {
    fn get_restrictions(&self, uid: u32) -> Result<Restrictions>;
}

#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize, zvariant::Type)]
pub struct Restriction {
    pub ip: IpAddr,
    pub hostname: String,
}

pub type Restrictions = Vec<Restriction>;

pub async fn restrictions_for(user: Uid) -> anyhow::Result<Vec<Restriction>, anyhow::Error> {
    #[cfg(not(feature = "integration_test"))]
    let proxy = {
        // This is the normal behavior
        let connection = zbus::Connection::system().await?;
        MalcontentDnsProxy::new(&connection).await?
    };

    #[cfg(feature = "integration_test")]
    let proxy = {
        // During integration testing, we want to connect to a private
        // bus name to avoid clashes with existing system services.
        let connection = zbus::Connection::session().await?;
        let dbus_name = std::env::var("TEST_DBUS_SERVICE_NAME")
            .expect("The test has not set the TEST_DBUS_SERVICE_NAME environment variable to the private bus name prior to attempting name resolution");
        MalcontentDnsProxy::builder(&connection)
            .destination(zbus_names::UniqueName::try_from(dbus_name).unwrap())
            .unwrap()
            .build()
            .await
            .expect("Unable to build DBus proxy object")
    };

    let restrictions = proxy.get_restrictions(user.as_raw()).await;
    log::trace!(
        "malcontent-nss: user {} restrictions are {:?}",
        user,
        &restrictions
    );

    Ok(restrictions?)
}
